/*
    comp7
    
    Class CompilerRuntime models the runtime system. Its method "error" 
    signals a error at runtime and terminates the program.
    
    Class Expr and its subclasses now have a method "eval" to evaluate the 
    expression. An expression can be a letter whose value was previously 
    assigned before the ':'  --- see the grammar. Then it is necessary to use
    a symbol table to keep the variables with its values. The Symbol Table is
    pointed by variable symbolTable.
         
    Grammar:
       Program ::= VarDecList ':' Expr   
       VarDecList ::=  | VarDec VarDecList 
       VarDec ::= Letter '=' Number
       Expr::=  '(' oper Expr  Expr ')' | Number | Letter
       Oper ::= '+' | '-' | '*' | '/'
       Number ::= '0'| '1' | ... | '9'
       Letter ::= 'A' | 'B'| ... | 'Z'| 'a'| 'b' | ... | 'z'
       
     
*/

import AST.*;
import java.util.*;
import java.lang.Character;


public class Compiler {

    public Program compile( char []p_input ) {
        input = p_input;
        tokenPos = 0;
        symbolTable = new Hashtable();
        nextToken();
        Program result = program();
        if ( token != '\0' )
          error("End of file expected");
        return result;
        }
    
    private Program program() {
        Vector arrayVariable = varDecList();
        if ( token != ':' ) {
            error(": expected");
            return null;
        }
        else {
            nextToken();
            Expr e = expr();
            return new Program( arrayVariable, e );
        }
    }

    private Vector varDecList() {
          /* See how the repetition in the grammar reflects in the code. Since VarDec
             always begin with a letter, if token is NOT a letter, then VarDecList is
             empty and null is returned */
        if ( ! Character.isLetter(token) )
          return null;
        else {
            Vector arrayVariable = new Vector();
            while ( Character.isLetter(token) ) 
              arrayVariable.addElement( varDec() );
            return arrayVariable;
            }
            
    }
    
    
    private Variable varDec() {
        char name = letter();
        if ( token != '=' ) {
          error("= expected");
          return null;
        }
        else {
          nextToken();
          NumberExpr n = number();
            // semantic analysis
            // inserts the variable in the Symbol Table. It will be used to
            // retrieve the value of the variable when it is found inside the
            // expression and to check if the variable found in the expression
            // was "declared" before the ':'
            // Note we insert the name converted to a String in the table
          if ( symbolTable.put( name + "", n ) != null ) 
            error("Variable " + name + " has already been declared");
          return new Variable( name, n.getValue() );
        }
    }

    private char letter() {
       if ( ! Character.isLetter(token) ) {
          error("Letter expected");
          return '\0';
       }
       else {
         char ch = token;
         nextToken();
         return ch;
       }
    }
    
    private Expr expr() {
        if ( token == '(' ) {
            nextToken();
            char op = oper();
            Expr e1 = expr();
            Expr e2 = expr();
            CompositeExpr ce = new CompositeExpr(e1, op, e2);
            if ( token == ')' ) 
              nextToken();
            else
              error(") expected");
            return ce;
        }
        else
            // note we test the token to decide which production to use
          if ( Character.isDigit(token) ) 
            return number();
          else
            return letterExpr();
    }
    
    
    private Expr letterExpr() {
        NumberExpr e;
        char ch = letter();
          // semantic analysis
          // was the variable declared ?
        e = (NumberExpr ) symbolTable.get( ch + "" );
        if ( e == null ) {
          error("Variable was not declared");
          return null;
        }
        else
          return new VariableExpr( ch, e.getValue() );
    }
    
    private NumberExpr number() {
        NumberExpr e = null;
        
        if ( token >= '0'  && token <= '9' ) {
            e = new NumberExpr(token);    
            nextToken();
        }
        else
          error("Digit expected");
        return e;
    }
    
    private char oper() {
        char op = token;
        if ( token == '+' || token == '-' || token == '*'  || token == '/' ) {
          nextToken();
          return op;
        }
        else {
          error("Operator expected");
            // should never execute since error terminates the program
          return '\0';
        }
    }
    
    private void nextToken() {
        while (  tokenPos < input.length && input[tokenPos] == ' ' ) 
          tokenPos++;
        if ( tokenPos < input.length ) {
          token = input[tokenPos];        
          tokenPos++;
        }
        else
          token = '\0';
    }
    
    private void error( String strMessage ) {
        if ( tokenPos == 0 ) 
          tokenPos = 1; 
        else 
          if ( tokenPos >= input.length )
            tokenPos = input.length;
        
        String strInput = new String( input, tokenPos - 1, input.length - tokenPos + 1 );
        String strError = "At \"" + strInput + "\"";
        System.out.print( strError );
        System.out.println( strMessage );
        throw new RuntimeException(strError);
    }
    
      // current token
    private char token;
    private int  tokenPos;
      // program given as input - source code
    private char []input;
    
    private Hashtable symbolTable;

        
}
